@@ -1,4 +1,4 @@
-use crate::constants::MAX_PRECISION;
+use crate::{constants::MAX_PRECISION_U32, Decimal};
 use alloc::string::String;
 use core::fmt;
 
@@ -21,6 +21,11 @@ where
     }
 }
 
+#[cold]
+pub(crate) fn tail_error(from: &'static str) -> Result<Decimal, Error> {
+    Err(from.into())
+}
+
 #[cfg(feature = "std")]
 impl std::error::Error for Error {}
 
@@ -38,7 +43,7 @@ impl fmt::Display for Error {
                 write!(
                     f,
                     "Scale exceeds the maximum precision allowed: {} > {}",
-                    scale, MAX_PRECISION
+                    scale, MAX_PRECISION_U32
                 )
             }
         }
