@@ -15,8 +15,8 @@
 //!
 //! ```toml
 //! [dependencies]
-//! rust_decimal = "1.19"
-//! rust_decimal_macros = "1.19"
+//! rust_decimal = "1.20"
+//! rust_decimal_macros = "1.20"
 //! ```
 //!
 //! ## Usage
@@ -190,7 +190,7 @@ mod postgres;
 #[cfg(feature = "rocket-traits")]
 mod rocket;
 #[cfg(feature = "serde")]
-mod serde;
+pub mod serde;
 
 pub use decimal::{Decimal, RoundingStrategy};
 pub use error::Error;
